#drinc:exec/memory.g
#drinc:disassemble.g
#drinc:util.g
#dis.g
#funcs.g

/*
 * dis.d - central symbolic disassembly routines.
 *
 * May/June 1989, by Chris Gray
 */

uint
    R_D0 = 0,
    R_D1 = 1,
    R_A0 = 0,
    R_FP = 6,

    OP_MOVEB = 0x1000,
    OP_MOVEL = 0x2000,

    M_DDIR = 0,
    M_ADIR = 1,
    M_DISP = 5;

Hunk_t TextHunk;
DisassemblerState_t Ds;
bool AbsoluteAddress, ShowAddresses, MakeLabels, WorkHard, IsLibrary,
    ForceDisAll, PutRelForm, PutPosition, CapExtended, DracoKludge,
    ForceKludge, PassOne;
ulong NewLabels;
ulong LibTableOffset;

/*
 * interface routines whereby the parameter parsing gives us our flags.
 */

proc setAbsoluteAddress(bool flag)void:

    AbsoluteAddress := flag;
corp;

proc setShowAddresses(bool flag)void:

    ShowAddresses := flag;
corp;

proc setMakeLabels(bool flag)void:

    MakeLabels := flag;
corp;

proc setWorkHard(bool flag)void:

    WorkHard := flag;
corp;

proc setIsLibrary(bool flag)void:

    IsLibrary := flag;
corp;

proc setForceDisAll(bool flag)void:

    ForceDisAll := flag;
corp;

proc setPutRelForm(bool flag)void:

    PutRelForm := flag;
corp;

proc setPutPosition(bool flag)void:

    PutPosition := flag;
corp;

proc setCapExtended(bool flag)void:

    CapExtended := flag;
corp;

proc setDracoKludge(bool flag)void:

    DracoKludge := flag;
corp;

proc setForceKludge(bool flag)void:

    ForceKludge := flag;
corp;

/*
 * readData - read in the contents of a data hunk.
 */

proc readData(register *Hunk_t h)void:
    register ulong length;
    register *byte data;

    length := readLong() * sizeof(ulong);
    data := AllocMem(length + 4, 0);
    h*.h_bytes := data;
    h*.h_bytesLength := length;
    if data ~= nil then
	while not getHadEof() and length ~= 0 do
	    length := length - sizeof(ulong);
	    data* := readByte();
	    data := data + 1;
	    data* := readByte();
	    data := data + 1;
	    data* := readByte();
	    data := data + 1;
	    data* := readByte();
	    data := data + 1;
	od;
    else
	skipLongs(length / sizeof(ulong));
    fi;
corp;

/*
 * putChar - library interface routine to output a character.
 */

proc putChar(/* char ch */)void:
    char ch;

    code(
	OP_MOVEB | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	ch
    );
    putChar1(ch);
corp;

/*
 * readTextWord - library interface routine to read a word from the input
 *	text hunk to disassemble.
 */

proc readTextWord(/* ulong address */)uint:
    ulong address;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	address
    );
    if address < TextHunk.h_bytesLength then
	(pretend(TextHunk.h_bytes, *uint) + address)*
    else
	0
    fi
corp;

/*
 * findLabel - library interface routine to check for a symbolic label at
 *	a given address in the code being disassembled.
 */

proc findLabel(/* ulong addr */)*char:
    ulong addr;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	addr
    );
    findLabel1(&TextHunk, addr)
corp;

/*
 * newBranch - a possible new branch within the text hunk.
 */

proc newBranch(ulong addr)void:

    if addr < TextHunk.h_bytesLength and not isBranch1(&TextHunk, addr) then
	branchTo1(&TextHunk, addr);
	NewLabels := NewLabels + 1;
    fi;
corp;

/*
 * findAbsSymbol - library interface routine to find any symbolic form for
 *	the address given, as used at the given offset.
 */

proc findAbsSymbol(/* ulong addr, refAt; *ulong pTrueAddr */)*char:
    ulong address, refAddr, hunkNumber;
    *ulong pTrueAddress;
    register ulong addr, refAt;
    register *char label;
    register *ulong pTrueAddr;
    register *DefList_t dfl;
    bool isBranch;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	address,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D1,
	refAddr,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	pTrueAddress
    );
    addr := address;
    refAt := refAddr;
    pTrueAddr := pTrueAddress;
    if addr & 0x80000000 ~= 0 then
	isBranch := true;
	addr := addr & 0x7fffffff;
    else
	isBranch := false;
    fi;
    label := findRelocOrDef(&TextHunk, addr, refAt, pTrueAddr, &hunkNumber);
    if label ~= nil then
	if hunkNumber = TextHunk.h_hunkNumber then
	    if TextHunk.h_labelsOK then
		if isBranch then
		    newBranch(addr);
		else
		    labelAt1(&TextHunk, addr);
		fi;
	    fi;
	elif hunkNumber > TextHunk.h_hunkNumber then
	    farLabel(addr, hunkNumber, isBranch);
	fi;
	label
    else
	dfl := TextHunk.h_absList;
	while dfl ~= nil and dfl*.dfl_offset ~= addr do
	    dfl := dfl*.dfl_next;
	od;
	if dfl ~= nil then
	    pTrueAddr* := dfl*.dfl_offset;
	    dfl*.dfl_name
	else
	    nil
	fi
    fi
corp;

/*
 * findRelCode - library interface routine to see if there is a symbolic name
 *	for a relative reference to code.
 */

proc findRelCode(/* long offset; ulong refAt */)*char:
    long offset;
    ulong refAt;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	offset,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D1,
	refAt
    );
    findRelocOrDef(&TextHunk, refAt + offset, refAt, &refAt, &refAt)
corp;

/*
 * findRelData - library interface routine to see if there is a symbolic name
 *	for a relative reference to data.
 */

proc findRelData(/* long offset; ulong refAt; *long pTrueOffset */)*char:
    ulong offset;
    ulong refAt;
    *ulong pTrueOffset;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	offset,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D1,
	refAt,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	pTrueOffset
    );
    findReloc(&TextHunk, offset, refAt, pTrueOffset, &offset)
corp;

/*
 * isLabel - library interface routine to see if the given offset has been
 *	used as a label.
 */

proc isLabel(/* ulong offset */)bool:
    ulong offset;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	offset
    );
    isLabel1(&TextHunk, offset)
corp;

/*
 * branchTo - library interface routine to mark an offset as a branch target.
 */

proc branchTo(/* ulong offset */)void:
    ulong offset;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	offset
    );
    if offset % 2 = 0 and offset < TextHunk.h_bytesLength then
	newBranch(offset);
    fi;
corp;

/*
 * labelAt - library interface routine to mark an offset as used as a label.
 */

proc labelAt(/* ulong offset */)void:
    ulong offset;

    code(
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	offset
    );
    if offset % 2 = 0 then
	labelAt1(&TextHunk, offset);
    fi;
corp;

/*
 * doPosition - utility routine to print a leading address portion.
 */

proc doPosition(ulong addr, base)void:

    if PutPosition then
	if AbsoluteAddress then
	    message("0x");
	    putLongHex(addr);
	else
	    message("   +0x");
	    putWordHex(addr - base);
	fi;
	putChar1(':');
    fi;
corp;

/*
 * buildLabel - build an appropriate label for 'findReloc'.
 */

proc buildLabel(register *char buffer; register ulong addr)void:
    register uint i, digit;

    if Ds.ds_absoluteAddress then
	i := 8;
    else
	addr := addr - Ds.ds_relativeBase;
	i := 4;
    fi;
    buffer* := 'L';
    buffer := buffer + (i + 1) * sizeof(char);
    buffer* := '\e';
    while i ~= 0 do
	i := i - 1;
	buffer := buffer - 1;
	digit := addr & 0xf;
	addr := addr >> 4;
	buffer* := if digit >= 10 then digit - 10 + 'a' else digit + '0' fi;
    od;
corp;

/*
 * putHeader - output the address and any needed label for a line of
 *	special disassembly output.
 */
	
proc putHeader()void:

    doPosition(Ds.ds_address, Ds.ds_relativeBase);
    if TextHunk.h_labelsOK and isLabel1(&TextHunk, Ds.ds_address) then
	putChar1('L');
	if AbsoluteAddress then
	    putLongHex(Ds.ds_address);
	    newLine();
	    doPosition(Ds.ds_address, Ds.ds_relativeBase);
	    message("        ");
	else
	    putWordHex(Ds.ds_address - Ds.ds_relativeBase);
	    message("   ");
	fi;
    else
	message("        ");
    fi;
corp;

/*
 * putLabel - output a use of a label.
 */

proc putLabel(bool isBranch)void:
    register *uint ptr;
    ulong trueAddr, hunkNumber, addr;

    putHeader();
    message("dc.l       ");
    ptr := TextHunk.h_bytes;
    ptr := ptr + Ds.ds_address;
    addr := make(ptr*, ulong) << 16 | make((ptr + 2)*, ulong);
    ignore findReloc(&TextHunk, addr, Ds.ds_address, &trueAddr, &hunkNumber);
    Ds.ds_address := Ds.ds_address + 4;
    if TextHunk.h_labelsOK then
	putChar1('L');
	if AbsoluteAddress then
	    putLongHex(addr);
	else
	    putWordHex(addr - Ds.ds_relativeBase);
	fi;
	if isBranch then
	    newBranch(addr);
	else
	    labelAt1(&TextHunk, addr);
	fi;
    else
	message("(reloc)0x");
	if AbsoluteAddress then
	    putLongHex(addr);
	else
	    putWordHex(addr - Ds.ds_relativeBase);
	fi;
    fi;
    newLine();
corp;

/*
 * doDracoCase - special handling for Draco's binary search case tables.
 */

proc doDracoCase(bool isLong)void:
    register *uint ptr;
    register uint count;

    ignore Disassemble(&Ds);
    putLabel(true);
    ptr := TextHunk.h_bytes;
    count := (ptr + Ds.ds_address)*;
    doPosition(Ds.ds_address, Ds.ds_relativeBase);
    message("        dc.w       ");
    putLongDec(count);
    newLine();
    Ds.ds_address := Ds.ds_address + 2;
    while count ~= 0 do
	count := count - 1;
	doPosition(Ds.ds_address, Ds.ds_relativeBase);
	message("        dc.");
	if isLong then
	    message("l       ");
	    putLongDec(make((ptr + Ds.ds_address)*, ulong) << 16 |
			   (ptr + Ds.ds_address + 2)*);
	    Ds.ds_address := Ds.ds_address + 4;
	else
	    message("w       ");
	    putLongDec(make((ptr + Ds.ds_address)*, uint));
	    Ds.ds_address := Ds.ds_address + 2;
	fi;
	newLine();
	putLabel(true);
    od;
corp;

/*
 * doOtherDracoCase - special handling for Draco's direct index case tables.
 */

proc doOtherDracoCase()void:
    register *uint ptr;
    register int difference;
    register ulong base, target;
    bool labelsOKSave;

    labelsOKSave := TextHunk.h_labelsOK;
    TextHunk.h_labelsOK := false;
    ignore Disassemble(&Ds);
    TextHunk.h_labelsOK := labelsOKSave;
    ptr := TextHunk.h_bytes;
    base := Ds.ds_address;
    while (ptr + Ds.ds_address)* & 0xf000 = 0xf000 do
	putHeader();
	message("dc.w       -");
	difference := (ptr + Ds.ds_address)*;
	putLongDec(- difference);
	if labelsOKSave then
	    target := base + difference - 2;
	    message("[L");
	    if AbsoluteAddress then
		putLongHex(target);
	    else
		putWordHex(target - Ds.ds_relativeBase);
	    fi;
	    putChar1(']');
	    newBranch(target);
	fi;
	newLine();
	Ds.ds_address := Ds.ds_address + 2;
    od;
corp;

/*
 * doLatticeCase - special handling for Lattice C's linear search switch tables
 */

proc doLatticeCase(bool isLong)void:
    register *uint ptr;
    register ulong offset;

    ignore Disassemble(&Ds);
    offset := if isLong then 4 else 2 fi;
    ptr := TextHunk.h_bytes;
    while (ptr + Ds.ds_address + offset)* & 0xff00 = 0x6000 do
	putHeader();
	if isLong then
	    message("dc.l       ");
	    putLongDec(make((ptr + Ds.ds_address)*, ulong) << 16 |
			    (ptr + Ds.ds_address + 2)*);
	    Ds.ds_address := Ds.ds_address + 4;
	else
	    message("dc.w       ");
	    putLongDec(make((ptr + Ds.ds_address)*, int));
	    Ds.ds_address := Ds.ds_address + 2;
	fi;
	newLine();
	ignore Disassemble(&Ds);
    od;
corp;

/*
 * doAztecCase - special handling for Aztec C's direct branch switch.
 *	NOTE: this will only work properly if 'WorkHard' is set.
 */

proc doAztecCase()void:
    register *uint ptr;
    register ulong base, addr;
    short offset;

    if TextHunk.h_labelsOK then
	ptr := TextHunk.h_bytes;
	base := Ds.ds_address;
	offset := (ptr + base - 2)* & 0x00ff;
	addr := base + offset - 2;
	while addr < base and not isBranch1(&TextHunk, addr) do
	    newBranch(base + make((ptr + addr)*, int) + 2);
	    addr := addr + 2;
	od;
    fi;
    ignore Disassemble(&Ds);
corp;

/*
 * doOtherCase - do another type of case that I noticed. (GreenHills?)
 */

proc doOtherCase(ulong length)void:
    register *uint ptr;
    register ulong base;
    register int offset;

    ptr := TextHunk.h_bytes;
    ignore Disassemble(&Ds);
    base := Ds.ds_address;
    /* not clear when to stop doing the labels with this one */
    while Ds.ds_address < length and not isBranch1(&TextHunk, Ds.ds_address) do
	offset := (ptr + Ds.ds_address)*;
	putHeader();
	message("dc.w       ");
	putLongDec(offset);
	if TextHunk.h_labelsOK then
	    message("[L");
	    if AbsoluteAddress then
		putLongHex(base + offset);
	    else
		putWordHex(base + offset - Ds.ds_relativeBase);
	    fi;
	    putChar1(']');
	    newBranch(base + offset);
	fi;
	newLine();
	Ds.ds_address := Ds.ds_address + 2;
    od;
corp;

/*
 * relocHere - see if the given name is referenced at the given location.
 */

proc relocHere(register *char name; register ulong refAt)bool:
    register *uint ptr;
    ulong addr, trueAddr, dummy;
    *char label;

    ptr := TextHunk.h_bytes;
    ptr := ptr + refAt;
    addr := make(ptr*, ulong) << 16 | make((ptr + 2)*, ulong);
    label := findReloc(&TextHunk, addr, refAt, &trueAddr, &dummy);
    label ~= nil and trueAddr = addr and CharsEqual(label, name)
corp;

/*
 * disassembleTo - disassemble up to a certain point in the text hunk.
 */

proc disassembleTo(ulong length; bool seenReturn)void:
    uint
	OP_SAVE1 = 0x2f00,
	OP_MOVEW_PC_INDEX = 0x303b,
	OP_SAVEM = 0x48e7,
	OP_LINK_A6 = 0x4e56,
	OP_RTS = 0x4e75, OP_RTD = 0x4e74,
	OP_JMP = 0x4ec0,
	OP_JMP_A0 = 0x4ed0, OP_JMP_A1 = 0x4ed1,
	OP_JMP_ABS_LONG = 0x4ef9, OP_JMP_PC_IND = 0x4efb,
	OP_BRA = 0x6000,
	IND_ZERO_D0 = 0x0000, IND_TWO_D0 = 0x0002,
	IND_SIX_D1 = 0x1006, IND_FOUR_D1 = 0x1004;
    register *uint ptr;
    register uint opCode, wrd;

    ptr := TextHunk.h_bytes;
    while not getInterrupted() and Ds.ds_address < length do
	while not getInterrupted() and Ds.ds_address < length and
	    (not seenReturn or ForceDisAll)
	do
	    opCode := (ptr + Ds.ds_address)*;
	    if Ds.ds_address >= length then
		wrd := 0x0000;
	    else
		wrd := (ptr + Ds.ds_address + 2)*;
	    fi;
	    if opCode = OP_JMP_ABS_LONG then
		if relocHere("_d_casew", Ds.ds_address + 2) then
		    doDracoCase(false);
		elif relocHere("_d_casel", Ds.ds_address + 2) then
		    doDracoCase(true);
		elif DracoKludge and Ds.ds_address + 20 < length then
		    opCode := (ptr + Ds.ds_address + 12)*;
		    wrd := (ptr + Ds.ds_address + 20)*;
		    if opCode = wrd and (wrd = 0 or wrd = 0xffff) then
			/* assume it is a long case */
			doDracoCase(true);
		    else
			doDracoCase(false);
		    fi;
		else
		    seenReturn := true;
		    ignore Disassemble(&Ds);
		fi;
	    elif opCode = OP_JMP_PC_IND and wrd = IND_ZERO_D0 and
		Ds.ds_address >= 4 and
		(ptr + Ds.ds_address - 4)* = OP_MOVEW_PC_INDEX
	    then
		if (ptr + Ds.ds_address - 2)* & 0x00ff >= 0x0080 then
		    doAztecCase();
		else
		    doOtherDracoCase();
		fi;
	    elif opCode = OP_JMP_PC_IND and wrd = IND_FOUR_D1 then
		doLatticeCase(false);
	    elif opCode = OP_JMP_PC_IND and wrd = IND_SIX_D1 then
		doLatticeCase(true);
	    elif opCode = OP_JMP_PC_IND and wrd = IND_TWO_D0 then
		doOtherCase(length);
	    else
		if (opCode = OP_RTS or opCode = OP_RTD or
			opCode = OP_JMP_A0 or opCode = OP_JMP_A1) and
		    (wrd ~= OP_SAVEM and wrd ~= OP_LINK_A6 and
			wrd & 0xfff0 ~= OP_SAVE1) or
		    (opCode & 0xff00 = OP_BRA or opCode & 0xffc0 = OP_JMP) and
		    TextHunk.h_labelsOK
		then
		    seenReturn := true;
		fi;
		ignore Disassemble(&Ds);
	    fi;
	od;
	if not getInterrupted() and Ds.ds_address < length then
	    if PassOne then
		skipData(&TextHunk, &Ds.ds_address, &Ds.ds_relativeBase,
			 length);
	    else
		showData(&TextHunk, &Ds.ds_address, &Ds.ds_relativeBase,
			 length, true, not AbsoluteAddress, AbsoluteAddress);
	    fi;
	fi;
	seenReturn := false;
    od;

    /* straight disassembly can overshoot */

    if Ds.ds_address > length then
	Ds.ds_address := length;
    fi;
corp;

/*
 * doLibrary - special handling if we are forced to treat this files as
 *	a library or device.
 */

proc doLibrary()void:
    uint OP_ILLEGAL = 0x4afc;
    register *uint ptr;
    register *ulong lPtr @ ptr;
    register *byte bPtr @ ptr;
    ulong initOffset, len, funcOffset;
    register uint i;
    bool startInData;

    ptr := TextHunk.h_bytes;

    if LibTableOffset = 0 then

	/* find the 'illegal' (RTC_MATCHWORD) to mark the special stuff */

	while not getInterrupted() and (ptr + Ds.ds_address)* ~= OP_ILLEGAL do
	    ignore Disassemble(&Ds);
	od;
	LibTableOffset := Ds.ds_address;

    else
	disassembleTo(LibTableOffset, false);
    fi;

    startInData := false;
    if (ptr + Ds.ds_address)* = OP_ILLEGAL then

	/* now we are at the library descriptor */

	putHeader();
	message("dc.w       0x4afc      ; RTC_MATCHWORD\n");
	Ds.ds_address := Ds.ds_address + 2;
	putLabel(false);
	putLabel(false);
	for i from 3 downto 0 do
	    putHeader();
	    message("dc.b       ");
	    putLongDec((bPtr + Ds.ds_address)*);
	    newLine();
	    Ds.ds_address := Ds.ds_address + 1;
	od;
	putLabel(false);
	putLabel(false);
	initOffset := (lPtr + Ds.ds_address)*;
	putLabel(false);

	len := TextHunk.h_bytesLength;
	if initOffset & 1 = 0 and initOffset < len then

	    /* scan forward for the init structure */

	    disassembleTo(initOffset, true);
	    if Ds.ds_address = initOffset and Ds.ds_address < len - 16 then

		/* There should be an init structure here, according to the
		   documentation, but some of the libraries and devices seem
		   to have just code there. Oh well. */

		if (lPtr + Ds.ds_address)* < 10000 then

		    /* the init structure */

		    putHeader();
		    message("dc.l       ");
		    putLongDec((lPtr + Ds.ds_address)*);
		    message("          ; positive library size\n");
		    Ds.ds_address := Ds.ds_address + 4;
		    funcOffset := (lPtr + Ds.ds_address)*;
		    putLabel(false);
		    putLabel(false);
		    putLabel(true);

		    if funcOffset & 1 = 0 and funcOffset < len then

			/* scan forward for the function table */
	
			disassembleTo(funcOffset, true);
			if Ds.ds_address = funcOffset then
	
			    /* the function table */
	
			    while Ds.ds_address < len - 4 and
				(lPtr + Ds.ds_address)* ~= 0xffffffff and
				not isBranch1(&TextHunk, Ds.ds_address) and
				not getInterrupted()
			    do
				putLabel(true);
			    od;
			    if Ds.ds_address < len - 4 then
				putHeader();
				message("dc.l       -1\n");
				Ds.ds_address := Ds.ds_address + 4;
				startInData := true;
			    fi;
			fi;
		    fi;
		fi;
	    fi;
	fi;
    fi;
    disassembleTo(TextHunk.h_bytesLength, startInData);
corp;

/*
 * doDisassembly - bottom level for one pass of disassembly of a hunk.
 */

proc doDisassembly()void:

    Ds.ds_readWord := readTextWord;
    Ds.ds_putChar := if PassOne then nil else putChar fi;
    Ds.ds_findLabel := if PassOne then nil else findLabel fi;
    Ds.ds_findAbsSymbol := findAbsSymbol;
    Ds.ds_findRelCode := findRelCode;
    Ds.ds_findRelData := findRelData;
    Ds.ds_labelAt :=
	if PassOne and TextHunk.h_labelsOK then labelAt else nil fi;
    Ds.ds_branchTo :=
	if PassOne and TextHunk.h_labelsOK then branchTo else nil fi;
    Ds.ds_isLabel :=
	if PassOne or not TextHunk.h_labelsOK then nil else isLabel fi;
    Ds.ds_address := 0;
    Ds.ds_relativeBase := 0;
    Ds.ds_operandColumn := if PutPosition then 20 + 11 else 20 fi;
    Ds.ds_putPosition := not PassOne and PutPosition;
    Ds.ds_absoluteAddress := AbsoluteAddress;
    Ds.ds_putErrors := true;
    Ds.ds_capExtended := CapExtended;
    Ds.ds_putAddress := ShowAddresses;
    Ds.ds_putRelForm := PutRelForm;
    if IsLibrary and TextHunk.h_hunkNumber = 0 then
	doLibrary();
    else
	disassembleTo(TextHunk.h_bytesLength, false);
    fi;
corp;

/*
 * dataHeader - output a header for a data or bss hunk.
 */

proc dataHeader(*Hunk_t h; bool isData)void:

    message("hunk_");
    message(if isData then "data" else "bss" fi);
    message(": (hunk ");
    putLongDec(h*.h_hunkNumber);
    message(", ");
    putLongDec(h*.h_bytesLength);
    message(" bytes):\n");
corp;

/*
 * doDataHunk - symbolic processing for a HUNK_DATA.
 */

proc doDataHunk(register *Hunk_t h)void:
    ulong offset, base;

    if h*.h_bytesLength ~= 0 and h*.h_bytes ~= nil then
	h*.h_labelsOK := false;
	if MakeLabels then
	    makeLabelMap(h);
	fi;
	if h*.h_labelsOK then
	    offset := 0;
	    base := 0;
	    setPassOne(true);
	    showData(h, &offset, &base, h*.h_bytesLength,
		     false, false, AbsoluteAddress);
	    h*.h_lastReloc := h*.h_relocEntries;
	    h*.h_lastLabel := h*.h_defList;
	fi;
	offset := 0;
	base := 0;
	setPassOne(false);
	showData(h, &offset, &base, h*.h_bytesLength,
		 false, false, AbsoluteAddress);
	if h*.h_labelsOK then
	    FreeMem(h*.h_branchMap, h*.h_mapLength);
	    FreeMem(h*.h_labelMap, h*.h_mapLength);
	fi;
    fi;
    freeSymbolic(h);
    if h*.h_bytes ~= nil then
	FreeMem(h*.h_bytes, h*.h_bytesLength + 4);
    fi;
corp;

/*
 * showCode - top level of disassembling a code hunk.
 */

proc showCode(*ulong pHunkNumber)ulong:
    Hunk_t dataHunk, bssHunk;
    Name_t name;
    register *uint p;
    register ulong i, labelLength;
    ulong nextHunk, hunkNumber, loop;
    bool gotData, gotBss, dataFirst, saveForceDisAll;

    hunkNumber := pHunkNumber*;
    TextHunk.h_hunkNumber := hunkNumber;
    hunkNumber := hunkNumber + 1;
    TextHunk.h_bytesLength := readLong() * sizeof(ulong);
    if TextHunk.h_bytesLength ~= 0 then
	p := AllocMem(TextHunk.h_bytesLength + 4, 0);
	if p = nil then
	    abort("can't allocate memory to buffer text");
	fi;
	TextHunk.h_bytes := p;
	i := TextHunk.h_bytesLength;
	while i ~= 0 do
	    i := i - 2;
	    p* := readWord();
	    p := p + sizeof(uint);
	od;
	TextHunk.h_labelsOK := false;
	if MakeLabels then
	    makeLabelMap(&TextHunk);
	fi;
	nextHunk := getSymbolic(&TextHunk);
	gotData := false;
	gotBss := false;
	while
	    while nextHunk = HUNK_END or nextHunk = HUNK_NAME do
		if nextHunk = HUNK_END then
		    nextHunk := readLong();
		else
		    readName(&name, readLong());
		    nextHunk := readLong();
		fi;
	    od;
	    nextHunk = HUNK_DATA and not gotData or
		nextHunk = HUNK_BSS and not gotBss
	do
	    if nextHunk = HUNK_DATA then
		dataHunk.h_hunkNumber := hunkNumber;
		hunkNumber := hunkNumber + 1;
		readData(&dataHunk);
		nextHunk := getSymbolic(&dataHunk);
		gotData := true;
		if not gotBss then
		    dataFirst := true;
		fi;
	    else
		bssHunk.h_hunkNumber := hunkNumber;
		hunkNumber := hunkNumber + 1;
		bssHunk.h_bytesLength := readLong() * sizeof(ulong);
		nextHunk := getSymbolic(&bssHunk);
		gotBss := true;
		if not gotData then
		    dataFirst := false;
		fi;
	    fi;
	od;
	if canDisassemble() then
	    LibTableOffset := 0;
	    if TextHunk.h_labelsOK then
		setPassOne(true);
		PassOne := true;
		if WorkHard then
		    saveForceDisAll := ForceDisAll;
		    if ForceKludge then
			ForceDisAll := true;
		    fi;
		    loop := 1;
		    while
			NewLabels := 0;
			doDisassembly();
			TextHunk.h_lastReloc := TextHunk.h_relocEntries;
			TextHunk.h_lastLabel := TextHunk.h_defList;
			not getInterrupted() and NewLabels ~= 0
		    do
			setPassOne(false);
			message("Loop ");
			putLongDec(loop);
			message(": ");
			putLongDec(NewLabels);
			message(" new labels.\n");
			setPassOne(true);
			loop := loop + 1;
		    od;
		    ForceDisAll := saveForceDisAll;
		else
		    doDisassembly();
		    TextHunk.h_lastReloc := TextHunk.h_relocEntries;
		    TextHunk.h_lastLabel := TextHunk.h_defList;
		fi;
	    fi;
	    setPassOne(false);
	    PassOne := false;
	    doDisassembly();
	fi;
	if gotData and dataFirst then
	    newLine();
	    dataHeader(&dataHunk, true);
	    doDataHunk(&dataHunk);
	    gotData := false;
	fi;
	if gotBss then
	    newLine();
	    dataHeader(&bssHunk, false);
	    if bssHunk.h_defList ~= nil then
		showBss(&bssHunk);
	    fi;
	fi;
	if gotData then
	    newLine();
	    dataHeader(&dataHunk, true);
	    doDataHunk(&dataHunk);
	fi;
	freeSymbolic(&TextHunk);
	if TextHunk.h_labelsOK then
	    FreeMem(TextHunk.h_branchMap, TextHunk.h_mapLength);
	    FreeMem(TextHunk.h_labelMap, TextHunk.h_mapLength);
	fi;
	FreeMem(TextHunk.h_bytes, TextHunk.h_bytesLength + 4);
    else
	nextHunk := HUNK_ILLEGAL;
    fi;
    pHunkNumber* := hunkNumber;
    nextHunk
corp;
