#dis.g
#funcs.g

/*
 * dump.d - routines for raw dump of object file.
 *
 * May/June 1989, by Chris Gray
 */

/*
 * dumpData - dump a data or other hunk as hex/ASCII.
 */

proc dumpData()void:
    [16] uint uintBuffer;
    [32] char charBuffer @ uintBuffer;
    [2] char buf;
    register ulong dataLen, pos, i, n;

    buf[1] := '\e';
    dataLen := readLong();
    pos := 0;
    while not getHadEof() and dataLen ~= 0 do
	n := 0;
	while not getHadEof() and dataLen ~= 0 and n ~= 8 do
	    uintBuffer[n] := readWord();
	    n := n + 1;
	    uintBuffer[n] := readWord();
	    n := n + 1;
	    dataLen := dataLen - 1;
	od;
	putLongHex(pos);
	message(": ");
	i := 0;
	while i ~= n do
	    putWordHex(uintBuffer[i]);
	    putChar1(' ');
	    i := i + 1;
	od;
	while i ~= 8 do
	    message("     ");
	    i := i + 1;
	od;
	putChar1('*');
	i := 0;
	n := n * 2;
	while i ~= n do
	    if charBuffer[i] < ' ' or charBuffer[i] > '~' then
		putChar1('.');
	    else
		buf[0] := charBuffer[i];
		message(&buf[0]);
	    fi;
	    i := i + 1;
	od;
	while i ~= 16 do
	    putChar1(' ');
	    i := i + 1;
	od;
	message("*\n");
	pos := pos + 16;
    od;
corp;

/*
 * indent - local routine for some formatting.
 */

proc indent()void:

    message("\n        ");
corp;

/*
 * dumpReloc - dump a set of relocation information.
 */

proc dumpReloc()void:
    register ulong n, hunk, count;

    message(":\n");
    while
	n := readLong();
	not getHadEof() and n ~= 0
    do
	hunk := readLong();
	if not getHadEof() then
	    message("    hunk ");
	    putLongDec(hunk);
	    putChar1(':');
	    count := 6;
	    while not getHadEof() and n ~= 0 do
		if count = 6 then
		    count := 0;
		    indent();
		fi;
		count := count + 1;
		putLongHex(readLong());
		putChar1(' ');
		n := n - 1;
	    od;
	    newLine();
	fi;
    od;
    newLine();
corp;

/*
 * skipReloc - skip a set of relocation information.
 */

proc skipReloc()void:
    register ulong n;

    newLine();
    while
	n := readLong();
	not getHadEof() and n ~= 0
    do
	ignore readLong();
	while not getHadEof() and n ~= 0 do
	    ignore readLong();
	    n := n - 1;
	od;
    od;
corp;

/*
 * dumpRefSet - dump the set of references to an external symbol, etc.
 */

proc dumpRefSet()void:
    register ulong n, count;

    n := readLong();
    if n = 0 then
	message(" no references\n");
    else
	count := 6;
	while not getHadEof() and n ~= 0 do
	    if count = 6 then
		count := 0;
		indent();
	    fi;
	    count := count + 1;
	    putLongHex(readLong());
	    putChar1(' ');
	    n := n - 1;
	od;
	newLine();
    fi;
corp;

/*
 * dumpExt - dump out a HUNK_EXT hunk.
 */

proc dumpExt()void:
    Name_t name;
    register ulong typeLen;

    message(":\n");
    while
	typeLen := readLong();
	not getHadEof() and typeLen ~= 0
    do
	readName(&name, typeLen & 0x00ffffff);
	message("    ");
	message(&name[0]);
	message(": ");
	case typeLen >> 24
	incase EXT_SYMB:
	    message("ext_symb: ");
	    putLongHex(readLong());
	    newLine();
	incase EXT_DEF:
	    message("ext_def: ");
	    putLongHex(readLong());
	    newLine();
	incase EXT_ABS:
	    message("ext_abs: ");
	    putLongHex(readLong());
	    newLine();
	incase EXT_RES:
	    message("ext_res: ");
	    putLongHex(readLong());
	    newLine();
	incase EXT_REF32:
	    message("ext_ref32:");
	    dumpRefSet();
	incase EXT_REF16:
	    message("ext_ref16:");
	    dumpRefSet();
	incase EXT_REF8:
	    message("ext_ref8:");
	    dumpRefSet();
	incase EXT_COMMON:
	    message("ext_common, len ");
	    putLongDec(readLong());
	    message(": ");
	    dumpRefSet();
	incase EXT_REF_REL32:
	    message("ext_ref_rel32:");
	    dumpRefSet();
	incase EXT_REF_REL16:
	    message("ext_ref_rel16:");
	    dumpRefSet();
	incase EXT_REF_REL8:
	    message("ext_ref_rel8:");
	    dumpRefSet();
	esac;
    od;
    newLine();
corp;

/*
 * skipExt - skip out a HUNK_EXT hunk.
 */

proc skipExt()void:
    Name_t name;
    register ulong typeLen;

    newLine();
    while
	typeLen := readLong();
	not getHadEof() and typeLen ~= 0
    do
	readName(&name, typeLen & 0x00ffffff);
	case typeLen >> 24
	incase EXT_SYMB:
	incase EXT_DEF:
	incase EXT_ABS:
	incase EXT_RES:
	    ignore readLong();
	incase EXT_REF32:
	incase EXT_REF16:
	incase EXT_REF8:
	incase EXT_REF_REL32:
	incase EXT_REF_REL16:
	incase EXT_REF_REL8:
	    skipLongs(readLong());
	incase EXT_COMMON:
	    ignore readLong();
	    skipLongs(readLong());
	esac;
    od;
corp;

/*
 * dumpSymbol - dump out a HUNK_SYMBOL hunk.
 */

proc dumpSymbol()void:
    Name_t name;
    register ulong nameLen;

    message(":\n");
    while
	nameLen := readLong();
	not getHadEof() and nameLen ~= 0
    do
	readName(&name, nameLen);
	message("    ");
	message(&name[0]);
	message(": ");
	putLongHex(readLong());
	newLine();
    od;
    newLine();
corp;

/*
 * skipSymbol - skip a HUNK_SYMBOL.
 */

proc skipSymbol()void:
    Name_t name;
    register ulong nameLen;

    newLine();
    while
	nameLen := readLong();
	not getHadEof() and nameLen ~= 0
    do
	readName(&name, nameLen);
	ignore readLong();
    od;
corp;

/*
 * dumpHeader - dump out a HUNK_HEADER hunk.
 */

proc dumpHeader()void:
    register ulong nameLen, first, last, count;
    Name_t name;
    bool hadLib;

    hadLib := false;
    while
	nameLen := readLong() & 0x00ffffff;
	not getHadEof() and nameLen ~= 0
    do
	hadLib := true;
	readName(&name, nameLen);
	message("    uses library ");
	message(&name[0]);
	newLine();
    od;
    if not hadLib then
	message(": ");
    fi;
    nameLen := readLong();
    first := readLong();
    last := readLong();
    message("table size: ");
    putLongDec(nameLen);
    message(", first hunk: ");
    putLongDec(first);
    message(", last hunk: ");
    putLongDec(last);
    message(", sizes:");
    last := last + 1;
    count := 6;
    while not getHadEof() and last ~= first do
	last := last - 1;
	if count = 6 then
	    count := 0;
	    indent();
	fi;
	count := count + 1;
	nameLen := readLong();
	putMemType(nameLen);
	putLongHex(nameLen & 0x00ffffff);
	putChar1(' ');
    od;
    newLine();
corp;

/*
 * skipHeader - skip over a HUNK_HEADER hunk.
 */

proc skipHeader()void:
    register ulong ul;
    Name_t name;

    while
	ul := readLong() & 0x00ffffff;
	not getHadEof() and ul ~= 0
    do
	readName(&name, ul);
    od;
    ignore readLong();
    ul := readLong();
    skipLongs(readLong() + 1 - ul);
corp;

/*
 * dumpOverlay - dump out a HUNK_OVERLAY hunk.
 */

proc dumpOverlay()void:
    register ulong size, mPlus2;

    size := readLong();
    mPlus2 := readLong();
    message(" size = ");
    putLongDec(size);
    message(", M + 2 = ");
    putLongDec(mPlus2);
    newLine();
    skipLongs(size);
corp;

/*
 * skipOverlay - skip a HUNK_OVERLAY.
 */

proc skipOverlay()void:
    register ulong n;

    n := readLong();
    ignore readLong();
    ignore readLong();
    skipLongs(n);
corp;
