#drinc:libraries/dos.g
#drinc:exec/miscellaneous.g
#drinc:exec/memory.g
#drinc:exec/tasks.g
#drinc:disassemble.g
#drinc:util.g
#dis.g
#funcs.g

/*
 * io.d - program startup and I/O portions of disassembler.
 *
 * May/June 1989, by Chris Gray
 */

uint BUFFER_SIZE = 512 * 20;
uint OUTPUT_SIZE = 100;

[BUFFER_SIZE] byte Buffer;
uint BufferMax, BufferPos;
[OUTPUT_SIZE] char OutputBuffer;
uint OutputPos;
Handle_t StdOut, Handle;
*char FileName;
bool HadEof, Interrupted;

*Library_t DisassembleBase;

bool HunksOnly, ShowAll, RawAll, PassOne;

/*
 * getInterrupted - tell the caller about having been interrupted.
 */

proc getInterrupted()bool:

    Interrupted
corp;

/*
 * getHadEof - tell the caller about having seen end-of-file.
 */

proc getHadEof()bool:

    HadEof
corp;

/*
 * setPassOne - allow higher level disassembly to control pass.
 */

proc setPassOne(bool flag)void:

    PassOne := flag;
corp;

/*
 * flushOutput - flush the output text buffer.
 */

proc flushOutput()void:

    if OutputPos ~= 0 then
	if not Interrupted then
	    ignore Write(StdOut, &OutputBuffer[0], OutputPos);
	fi;
	OutputPos := 0;
    fi;
    if SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C ~= 0 then
	if not Interrupted then
	    ignore Write(StdOut, "Interrupt! - Dis exiting.\n", 26);
	fi;
	Interrupted := true;
	HadEof := true;
    fi;
corp;

/*
 * putChar1 - bottom level routine to output a single character.
 */

proc putChar1(char ch)void:

    if not PassOne then
	if OutputPos = OUTPUT_SIZE then
	    flushOutput();
	fi;
	OutputBuffer[OutputPos] := ch;
	OutputPos := OutputPos + 1;
	if ch = '\n' then
	    flushOutput();
	fi;
    fi;
corp;

/*
 * newLine - output a '\n'.
 */

proc newLine()void:

    putChar1('\n');
corp;

/*
 * spaceDown - do an extra line if not HunksOnly.
 */

proc spaceDown()void:

    if not HunksOnly then
	putChar1('\n');
    fi;
corp;

/*
 * message - output a simple string to the buffered output.
 */

proc message(register *char m)void:

    while m* ~= '\e' do
	putChar1(m*);
	m := m + sizeof(char);
    od;
corp;

/*
 * err - output a simple one-string error message.
 */

proc err(*char m)void:

    message("*** ");
    message(m);
    message(" ***\n");
corp;

/*
 * fileError - output an error message involving the file name.
 */

proc fileError(*char m)void:

    message(FileName);
    message(": ");
    err(m);
corp;

/*
 * abort - print message and exit
 */

proc abort(*char m)void:

    fileError(m);
    if DisassembleBase ~= nil then
	CloseDisassembleLibrary();
    fi;
    Close(Handle);
    Exit(10);
corp;

/*
 * readByte - read a byte from the current input object file.
 */

proc readByte()byte:

    if BufferPos = BufferMax then
	BufferMax := Read(Handle, &Buffer[0], BUFFER_SIZE);
	if SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C ~= 0 then
	    if not Interrupted then
		ignore Write(StdOut, "Interrupt! - Dis exiting.\n", 26);
	    fi;
	    Interrupted := true;
	    HadEof := true;
	fi;
	if BufferMax ~= 0 then
	    BufferPos := 1;
	    Buffer[0]
	else
	    HadEof := true;
	    BufferPos := 0;
	    0
	fi
    else
	BufferPos := BufferPos + 1;
	Buffer[BufferPos - 1]
    fi
corp;

/*
 * readWord - read a 16 bit word from the current input object file.
 */

proc readWord()uint:

    make(readByte(), uint) << 8 | make(readByte(), uint)
corp;

/*
 * readLong - read a 32 bit longword from the current input object file.
 */

proc readLong()ulong:

    make(readWord(), ulong) << 16 | make(readWord(), ulong)
corp;

/*
 * readName - read a hunk, etc. name from the input object file.
 */

proc readName(*Name_t name; register ulong nameLen)void:
    register ulong i;

    i := 0;
    while not HadEof and i ~= NAME_MAX - 1 and nameLen ~= 0 do
	name*[i] := readByte() + '\e';
	i := i + 1;
	if i % 4 = 0 then
	    nameLen := nameLen - 1;
	fi;
    od;
    name*[i] := '\e';
corp;

/*
 * skipLongs - skip the indicated number of longwords in the input file.
 */

proc skipLongs(register ulong n)void:

    while not HadEof and n ~= 0 do
	n := n - 1;
	ignore readLong();
    od;
corp;

/*
 * putWordHex - output a 16 bit value in hexadecimal.
 */

proc putWordHex(register uint wrd)void:
    [5] char buffer;
    register *char p;
    register uint i, digit;

    p := &buffer[4];
    p* := '\e';
    for i from 3 downto 0 do
	p := p - sizeof(char);
	digit := wrd & 0xf;
	p* := if digit > 9 then digit - 10 + 'a' else digit + '0' fi;
	wrd := wrd >> 4;
    od;
    message(p);
corp;

/*
 * putLongHex - output a 32 bit longword in hexadecimal.
 */

proc putLongHex(ulong wrd)void:

    putWordHex(wrd >> 16);
    putWordHex(wrd);
corp;

/*
 * putLongDec - output a 32 bit longword in decimal.
 */

proc putLongDec(ulong wrd)void:
    [11] char buffer;
    register *char p;

    p := &buffer[10];
    p* := '\e';
    while
	p := p - sizeof(char);
	p* := wrd % 10 + '0';
	wrd := wrd / 10;
	wrd ~= 0
    do
    od;
    message(p);
corp;

/*
 * putMemType - utility routine to output an AmigaDOS memory type.
 */

proc putMemType(ulong typ)void:

    case typ >> 24
    incase 0x80:
	message("fast ");
    incase 0x40:
	message("chip ");
    incase 0xc0:
	message("fast-chip?? ");
    esac;
corp;

/*
 * canDisassemble - try to open disassemble.library
 */

proc canDisassemble()bool:

    if DisassembleBase = nil then
	DisassembleBase := OpenDisassembleLibrary(0);
	if DisassembleBase = nil then
	    message("Can't open 'disassemble.library'\n");
	fi;
    fi;
    DisassembleBase ~= nil
corp;

/*
 * disassembleFile - top level of disassembling or dumping the set up file.
 */

proc disassembleFile()void:
    Hunk_t hunk;
    Name_t name;
    register ulong nextHunkType, thisHunkType;
    ulong hunkNumber, offset, base, length;
    bool eofOk;
		
    symbolInit();
    hunkNumber := 0;
    nextHunkType := HUNK_ILLEGAL;
    while
	if nextHunkType = HUNK_ILLEGAL then
	    thisHunkType := readLong();
	else
	    thisHunkType := nextHunkType;
	    nextHunkType := HUNK_ILLEGAL;
	fi;
	not HadEof
    do
	eofOk := false;
	putMemType(thisHunkType);
	case thisHunkType & 0x00ffffff
	incase HUNK_UNIT:
	    readName(&name, readLong());
	    if ShowAll then
		message("hunk_unit: ");
		message(&name[0]);
		newLine();
		spaceDown();
	    else
		if name[0] ~= '\e' then
		    message(&name[0]);
		    message(":\n\n");
		fi;
	    fi;
	incase HUNK_NAME:
	    readName(&name, readLong());
	    if ShowAll then
		message("hunk_name: ");
		message(&name[0]);
		newLine();
		spaceDown();
	    fi;
	incase HUNK_CODE:
	    message("hunk_code (hunk ");
	    putLongDec(hunkNumber);
	    if HunksOnly then
		length := readLong();
		skipLongs(length);
		message(", ");
		putLongDec(length * sizeof(ulong));
		message(" bytes)\n");
		hunkNumber := hunkNumber + 1;
	    else
		message("):\n");
		if RawAll then
		    dumpData();
		    hunkNumber := hunkNumber + 1;
		else
		    nextHunkType := showCode(&hunkNumber);
		    eofOk := true;
		fi;
		newLine();
	    fi;
	incase HUNK_DATA:
	    message("hunk_data (hunk ");
	    putLongDec(hunkNumber);
	    if HunksOnly then
		length := readLong();
		skipLongs(length);
		message(", ");
		putLongDec(length * sizeof(ulong));
		message(" bytes)\n");
	    else
		message("):\n");
		if RawAll then
		    dumpData();
		else
		    hunk.h_hunkNumber := hunkNumber;
		    readData(&hunk);
		    nextHunkType := getSymbolic(&hunk);
		    doDataHunk(&hunk);
		fi;
		newLine();
	    fi;
	    hunkNumber := hunkNumber + 1;
	incase HUNK_BSS:
	    length := readLong() * sizeof(ulong);
	    message("hunk_bss (hunk ");
	    putLongDec(hunkNumber);
	    message(", ");
	    putLongDec(length);
	    message(" bytes)");
	    if not HunksOnly and not RawAll then
		hunk.h_hunkNumber := hunkNumber;
		nextHunkType := getSymbolic(&hunk);
		if hunk.h_defList ~= nil then
		    hunk.h_bytesLength := length;
		    showBss(&hunk);
		fi;
	    fi;
	    newLine();
	    hunkNumber := hunkNumber + 1;
	incase HUNK_RELOC32:
	    message("hunk_reloc32");
	    if HunksOnly then
		skipReloc();
	    else
		dumpReloc();
	    fi;
	incase HUNK_RELOC16:
	    message("hunk_reloc16");
	    if HunksOnly then
		skipReloc();
	    else
		dumpReloc();
	    fi;
	incase HUNK_RELOC8:
	    message("hunk_reloc8");
	    if HunksOnly then
		skipReloc();
	    else
		dumpReloc();
	    fi;
	incase HUNK_EXT:
	    message("hunk_ext");
	    if HunksOnly then
		skipExt();
	    else
		dumpExt();
	    fi;
	incase HUNK_SYMBOL:
	    message("hunk_symbol");
	    if HunksOnly then
		skipSymbol();
	    else
		dumpSymbol();
	    fi;
	incase HUNK_DEBUG:
	    if ShowAll then
		message("hunk_debug");
		if HunksOnly then
		    skipLongs(readLong());
		else
		    message(":\n");
		    dumpData();
		fi;
		newLine();
	    else
		skipLongs(readLong());
	    fi;
	incase HUNK_END:
	    if ShowAll then
		message("hunk_end\n");
		spaceDown();
	    fi;
	incase HUNK_HEADER:
	    if ShowAll then
		message("hunk_header");
		if HunksOnly then
		    skipHeader();
		else
		    dumpHeader();
		fi;
		newLine();
	    else
		skipHeader();
	    fi;
	incase HUNK_OVERLAY:
	    message("hunk_overlay");
	    if HunksOnly then
		skipOverlay();
	    else
		dumpOverlay();
	    fi;
	    newLine();
	incase HUNK_BREAK:
	    if ShowAll then
		message("hunk_break\n");
	    fi;
	incase HUNK_RELOC_REL32:
	    message("hunk_reloc_rel32");
	    if HunksOnly then
		skipReloc();
	    else
		dumpReloc();
	    fi;
	incase HUNK_RELOC_REL16:
	    message("hunk_reloc_rel16");
	    if HunksOnly then
		skipReloc();
	    else
		dumpReloc();
	    fi;
	incase HUNK_RELOC_REL8:
	    message("hunk_reloc_rel8");
	    if HunksOnly then
		skipReloc();
	    else
		dumpReloc();
	    fi;
	incase HUNK_LIB:
	    message("hunk_lib ");
	    putLongDec(readLong());
	    newLine();
	    spaceDown();
	incase HUNK_DIRECTORY:
	    message("hunk_directory ");
	    length := readLong();
	    putLongDec(length);
	    newLine();
	    skipLongs(length);
	    spaceDown();
	default:
	    message(FileName);
	    message(": *** unknown hunk type ");
	    putLongHex(thisHunkType);
	    message(" ***\n");
	esac;
	if HadEof and not Interrupted and not eofOk then
	    fileError("unexpected end-of-file");
	fi;
    od;
    symbolTerm();
corp;

/*
 * main - parameter handling, file set-up, etc.
 */

proc main()void:
    *char USAGE = "Use is: dis [-hxasfnwlrpcqdk?] file1 ... fileN\n";
    extern
	_d_pars_initialize()void;
    bool hadFile, hadError;

    if OpenExecLibrary(0) ~= nil then
	if OpenDosLibrary(0) ~= nil then
	    DisassembleBase := nil;
	    StdOut := Output();
	    _d_pars_initialize();
	    Interrupted := false;
	    PassOne := false;
	    HunksOnly := false;
	    ShowAll := false;
	    RawAll := false;
	    setAbsoluteAddress(false);
	    setShowAddresses(false);
	    setMakeLabels(true);
	    setWorkHard(true);
	    setIsLibrary(false);
	    setForceDisAll(false);
	    setPutRelForm(false);
	    setPutPosition(true);
	    setCapExtended(true);
	    setDracoKludge(false);
	    setForceKludge(false);
	    hadFile := false;
	    hadError := false;
	    while
		FileName := GetPar();
		not Interrupted and not hadError and FileName ~= nil
	    do
		if FileName* = '-' then
		    while
			FileName := FileName + sizeof(char);
			FileName* ~= '\e'
		    do
			case FileName*
			incase 'h':
			incase 'H':
			    ShowAll := true;
			incase 'x':
			incase 'X':
			    ShowAll := true;
			    RawAll := true;
			incase 'a':
			incase 'A':
			    setAbsoluteAddress(true);
			incase 's':
			incase 'S':
			    setShowAddresses(true);
			incase 'f':
			incase 'F':
			    setForceDisAll(true);
			incase 'n':
			incase 'N':
			    setMakeLabels(false);
			incase 'w':
			incase 'W':
			    setWorkHard(false);
			incase 'l':
			incase 'L':
			    setIsLibrary(true);
			incase 'r':
			incase 'R':
			    setPutRelForm(true);
			incase 'p':
			incase 'P':
			    setPutPosition(false);
			incase 'c':
			incase 'C':
			    setCapExtended(false);
			incase 'q':
			incase 'Q':
			    HunksOnly := true;
			    ShowAll := true;
			incase 'd':
			incase 'D':
			    setDracoKludge(true);
			incase 'k':
			incase 'K':
			    setForceKludge(true);
			incase '?':
			    message(
"Dis V1.0, by Chris Gray\n"
"   flags:   -h - show all hunks\n"
"            -x - raw dump of all hunks\n"
"            -a - use absolute (hunk-relative) addresses\n"
"            -s - show hex address on branches, jumps, etc.\n"
"            -f - force disassembly of all code\n"
"            -n - do not generate labels for branch targets\n"
"            -w - disable hard work to distinguish code/data\n"
"            -l - treat file as a library/device\n"
"            -r - show relative form as well as label/address\n"
"            -p - don't output offsets at left of listing\n"
"            -c - don't capitalize non-68000 stuff\n"
"            -q - show summary of hunks only\n"
"            -d - kludge for Draco cases when fully linked\n"
"            -k - kludge to force code on pre-passes\n"
			    );
			    hadFile := true;
			default:
			    message(USAGE);
			    hadError := true;
			esac;
		    od;
		else
		    Handle := Open(FileName, MODE_OLDFILE);
		    if Handle = 0 then
			fileError("can't open");
		    else
			BufferMax := 0;
			BufferPos := 0;
			HadEof := false;
			OutputPos := 0;
			disassembleFile();
			Close(Handle);
		    fi;
		    hadFile := true;
		fi;
		flushOutput();
	    od;
	    if not hadFile and not hadError then
		message(USAGE);
		Exit(5);
	    fi;
	    flushOutput();
	    if DisassembleBase ~= nil then
		CloseDisassembleLibrary();
	    fi;
	    CloseDosLibrary();
	fi;
	CloseExecLibrary();
    fi;
corp;
