#drinc:exec/memory.g
#drinc:util.g
#dis.g
#funcs.g

/*
 * symbol.d - symbolic information manipulation routines.
 *
 * May/June 1989, by Chris Gray
 */

bool DEBUG = false;

*HunkList_t HunkList;
RelocEntry_t FakeReloc;
[17] char HunkNameBuffer;

/*
 * sortDefList - sort the indicated list of definitions by address.
 */

proc sortDefList(**DefList_t pDefList)void:
    register *DefList_t dfl, dflTemp;
    register **DefList_t pdfl;
    *DefList_t newHead;

    newHead := nil;
    dfl := pDefList*;
    while dfl ~= nil do
	pdfl := &newHead;
	while pdfl* ~= nil and pdfl**.dfl_offset < dfl*.dfl_offset do
	    pdfl := &pdfl**.dfl_next;
	od;
	dflTemp := dfl;
	dfl := dfl*.dfl_next;
	dflTemp*.dfl_next := pdfl*;
	pdfl* := dflTemp;
    od;
    pDefList* := newHead;
corp;

/*
 * insertRelocs - insert a set of relocation references into the set being
 *	built up globally for this hunk.
 */

proc insertRelocs(*Hunk_t h; *ulong refArray; ulong refCount;
		  register bool isRef; register arbptr symbol)void:
    register *ulong refs;
    register *RelocEntry_t rle;
    register ulong count1, count2, addr;

    count1 := refCount;
    refs := refArray;
    while count1 ~= 0 do
	count1 := count1 - 1;
	addr := refs*;
	refs := refs + sizeof(ulong);
	rle := h*.h_relocEntries;
	count2 := h*.h_relocCount;
	while count2 ~= 0 and addr > rle*.rle_offset do
	    count2 := count2 - 1;
	    rle := rle + sizeof(RelocEntry_t);
	od;
	if count2 ~= 0 then
	    BlockCopyB(rle + (count2 + 1) * sizeof(RelocEntry_t) - 1,
		       rle + count2 * sizeof(RelocEntry_t) - 1,
		       count2 * sizeof(RelocEntry_t));
	fi;
	rle*.rle_offset := addr;
	rle*.rle_isRef := isRef;
	rle*.rle_u.rle_hunk := symbol;
	h*.h_relocCount := h*.h_relocCount + 1;
    od;
    FreeMem(refArray, refCount * sizeof(ulong));
corp;

/*
 * mergeRelocs - merge all the relocation information for the current hunk.
 */

proc mergeRelocs(register *Hunk_t h)void:
    register *RelocList_t rll;
    register *HunkList_t hkl;
    register *RefList_t rfl @ rll;
    register *RelocEntry_t rle @ rll;
    register ulong hunk;

    h*.h_relocCount := 0;
    rll := h*.h_relocList;
    while rll ~= nil do
	hunk := rll*.rll_hunkNumber;
	hkl := HunkList;
	while hkl ~= nil and hkl*.hkl_hunkNumber ~= hunk do
	    hkl := hkl*.hkl_next;
	od;
	if hkl = nil then
	    hkl := AllocMem(sizeof(HunkList_t), 0);
	    if hkl ~= nil then
		hkl*.hkl_next := HunkList;
		hkl*.hkl_symbols := nil;
		hkl*.hkl_labelList := nil;
		hkl*.hkl_hunkNumber := hunk;
		HunkList := hkl;
	    fi;
	fi;
	insertRelocs(h, rll*.rll_refs, rll*.rll_count, false, hkl);
	rll := rll*.rll_next;
    od;
    rfl := h*.h_refList;
    while rfl ~= nil do
	insertRelocs(h, rfl*.rfl_refs, rfl*.rfl_count, true, rfl);
	rfl := rfl*.rfl_next;
    od;
    if DEBUG then
	rle := h*.h_relocEntries;
	for hunk from 1 upto h*.h_relocCount do
	    message("ref to ");
	    if rle*.rle_u.rle_ref ~= nil then
		if rle*.rle_isRef then
		    message(rle*.rle_u.rle_ref*.rfl_name);
		else
		    message("hunk ");
		    putLongHex(rle*.rle_u.rle_hunk*.hkl_hunkNumber);
		fi;
	    fi;
	    message(" at ");
	    putLongHex(rle*.rle_offset);
	    newLine();
	    rle := rle + sizeof(RelocEntry_t);
	od;
    fi;
    rle := h*.h_relocEntries + h*.h_relocCount * sizeof(RelocEntry_t);
    rle*.rle_offset := 0xffffffff;
    rle*.rle_u.rle_hunk := nil;
    rle*.rle_isRef := false;
    h*.h_relocCount := h*.h_relocCount + 1;
    h*.h_lastReloc := h*.h_relocEntries;
corp;

/*
 * getSymbolic - read a set of symbolic information for the current hunk
 *	from the input object file. Prepare the information for use during
 *	disassembly.
 */

proc getSymbolic(register *Hunk_t h)ulong:
    register *DefList_t dfl;
    register *RefList_t rfl @ dfl;
    register *RelocList_t rll @ dfl;
    register *ulong refs;
    register *HunkList_t hkl @ dfl;
    register ulong ul, relocCount, hunkType;
    Name_t name;

    h*.h_defList := nil;
    h*.h_absList := nil;
    h*.h_refList := nil;
    h*.h_relocList := nil;
    relocCount := 0;
    while
	hunkType := readLong();
	ul := hunkType & 0x00ffffff;
	ul = HUNK_EXT or ul = HUNK_RELOC32 or ul = HUNK_RELOC16 or
	    ul = HUNK_RELOC8 or ul = HUNK_SYMBOL or ul = HUNK_RELOC_REL32 or
	    ul = HUNK_RELOC_REL16 or ul = HUNK_RELOC_REL8
    do
	case ul
	incase HUNK_EXT:
	    while
		ul := readLong();
		not getHadEof() and ul ~= 0
	    do
		readName(&name, ul & 0x00ffffff);
		case ul >> 24
		incase EXT_DEF:
		    dfl := AllocMem(sizeof(DefList_t), 0);
		    if dfl ~= nil then
			dfl*.dfl_next := h*.h_defList;
			dfl*.dfl_name := AllocMem(CharsLen(&name[0]) + 1, 0);
			if dfl*.dfl_name ~= nil then
			    CharsCopy(dfl*.dfl_name, &name[0]);
			    dfl*.dfl_offset := readLong();
			    h*.h_defList := dfl;
			else
			    ignore readLong();
			    FreeMem(dfl, sizeof(DefList_t));
			fi;
		    else
			ignore readLong();
		    fi;
		incase EXT_ABS:
		    dfl := AllocMem(sizeof(DefList_t), 0);
		    if dfl ~= nil then
			dfl*.dfl_next := h*.h_absList;
			dfl*.dfl_name := AllocMem(CharsLen(&name[0]) + 1, 0);
			if dfl*.dfl_name ~= nil then
			    CharsCopy(dfl*.dfl_name, &name[0]);
			    dfl*.dfl_offset := readLong();
			    h*.h_absList := dfl;
			else
			    ignore readLong();
			    FreeMem(dfl, sizeof(DefList_t));
			fi;
		    else
			ignore readLong();
		    fi;
		incase EXT_REF32:
		incase EXT_REF16:
		incase EXT_REF8:
		incase EXT_REF_REL32:
		incase EXT_REF_REL16:
		incase EXT_REF_REL8:
		    rfl := AllocMem(sizeof(RefList_t), 0);
		    if rfl ~= nil then
			rfl*.rfl_next := h*.h_refList;
			rfl*.rfl_refs := nil;
			rfl*.rfl_name := AllocMem(CharsLen(&name[0]) + 1, 0);
			if rfl*.rfl_name ~= nil then
			    CharsCopy(rfl*.rfl_name, &name[0]);
			    ul := readLong();
			    refs := AllocMem(ul * sizeof(ulong), 0);
			    if refs ~= nil then
				relocCount := relocCount + ul;
				rfl*.rfl_refs := refs;
				rfl*.rfl_count := ul;
				while ul ~= 0 and not getHadEof() do
				    ul := ul - 1;
				    refs* := readLong();
				    refs := refs + sizeof(ulong);
				od;
				h*.h_refList := rfl;
			    else
				skipLongs(ul);
				FreeMem(rfl*.rfl_name, CharsLen(&name[0]) + 1);
				FreeMem(rfl, sizeof(RefList_t));
			    fi;
			else
			    skipLongs(readLong());
			    FreeMem(rfl, sizeof(RefList_t));
			fi;
		    else
			skipLongs(readLong());
		    fi;
		esac;
	    od;
	incase HUNK_RELOC32:
	incase HUNK_RELOC16:
	incase HUNK_RELOC8:
	incase HUNK_RELOC_REL32:
	incase HUNK_RELOC_REL16:
	incase HUNK_RELOC_REL8:
	    while
		ul := readLong();
		not getHadEof() and ul ~= 0
	    do
		rll := AllocMem(sizeof(RelocList_t), 0);
		if rll ~= nil then
		    rll*.rll_next := h*.h_relocList;
		    rll*.rll_hunkNumber := readLong();
		    refs := AllocMem(ul * sizeof(ulong), 0);
		    if refs ~= nil then
			relocCount := relocCount + ul;
			rll*.rll_refs := refs;
			rll*.rll_count := ul;
			while ul ~= 0 and not getHadEof() do
			    ul := ul - 1;
			    refs* := readLong();
			    refs := refs + sizeof(ulong);
			od;
			h*.h_relocList := rll;
		    else
			FreeMem(rll, sizeof(RelocList_t));
			skipLongs(ul);
		    fi;
		else
		    ignore readLong();
		    skipLongs(ul);
		fi;
	    od;
	incase HUNK_SYMBOL:
	    while
		ul := readLong();
		not getHadEof() and ul ~= 0
	    do
		readName(&name, ul);
		dfl := AllocMem(sizeof(DefList_t), 0);
		if dfl ~= nil then
		    dfl*.dfl_next := h*.h_defList;
		    dfl*.dfl_name := AllocMem(CharsLen(&name[0]) + 1, 0);
		    if dfl*.dfl_name ~= nil then
			CharsCopy(dfl*.dfl_name, &name[0]);
			dfl*.dfl_offset := readLong();
			h*.h_defList := dfl;
		    else
			ignore readLong();
			FreeMem(dfl, sizeof(DefList_t));
		    fi;
		else
		    ignore readLong();
		fi;
	    od;
	esac;
    od;
    sortDefList(&h*.h_defList);
    sortDefList(&h*.h_absList);
    h*.h_lastLabel := h*.h_defList;
    if DEBUG then
	message("Defined symbols in hunk ");
	putLongDec(h*.h_hunkNumber);
	message(":\n");
	dfl := h*.h_defList;
	while dfl ~= nil do
	    message(dfl*.dfl_name);
	    message(" at ");
	    putLongHex(dfl*.dfl_offset);
	    newLine();
	    dfl := dfl*.dfl_next;
	od;
    fi;
    if h*.h_defList ~= nil then
	hkl := HunkList;
	while hkl ~= nil and hkl*.hkl_hunkNumber ~= h*.h_hunkNumber do
	    hkl := hkl*.hkl_next;
	od;
	if hkl = nil then
	    hkl := AllocMem(sizeof(HunkList_t), 0);
	    if hkl ~= nil then
		hkl*.hkl_next := HunkList;
		hkl*.hkl_labelList := nil;
		hkl*.hkl_hunkNumber := h*.h_hunkNumber;
		HunkList := hkl;
	    fi;
	fi;
	if hkl ~= nil then
	    hkl*.hkl_symbols := h*.h_defList;
	fi;
    fi;
    h*.h_relocEntries := AllocMem((relocCount + 1) * sizeof(RelocEntry_t), 0);
    if h*.h_relocEntries ~= nil then
	mergeRelocs(h);
    else
	h*.h_relocCount := 0;
	h*.h_relocEntries := &FakeReloc;
    fi;
    hunkType
corp;

/*
 * freeDefList - utility routine to free a list of DefList_t's.
 */

proc freeDefList(register *DefList_t dfl)void:
    register *DefList_t dflTemp;

    while dfl ~= nil do
	FreeMem(dfl*.dfl_name, CharsLen(dfl*.dfl_name) + 1);
	dflTemp := dfl;
	dfl := dfl*.dfl_next;
	FreeMem(dflTemp, sizeof(DefList_t));
    od;
corp;

/*
 * freeLabelList - utility routine to free a list of LabelList_t's.
 */

proc freeLabelList(register *LabelList_t lal)void:
    register *LabelList_t lalTemp;

    while lal ~= nil do
	lalTemp := lal;
	lal := lal*.lal_next;
	FreeMem(lalTemp, sizeof(LabelList_t));
    od;
corp;

/*
 * freeSymbolic - free the symbolic and relocation information associated
 *	with the current hunk. Symbols defined within the hunk are kept,
 *	for use when disassembling later hunks which may reference this one.
 */

proc freeSymbolic(register *Hunk_t h)void:
    register *RefList_t rfl, rflTemp;
    register *RelocList_t rll @ rfl, rllTemp @ rflTemp;

    freeDefList(h*.h_absList);
    rfl := h*.h_refList;
    while rfl ~= nil do
	FreeMem(rfl*.rfl_name, CharsLen(rfl*.rfl_name) + 1);
	rflTemp := rfl;
	rfl := rfl*.rfl_next;
	FreeMem(rflTemp, sizeof(RefList_t));
    od;
    rll := h*.h_relocList;
    while rll ~= nil do
	rllTemp := rll;
	rll := rll*.rll_next;
	FreeMem(rllTemp, sizeof(RelocList_t));
    od;
    if h*.h_relocCount ~= 0 then
	FreeMem(h*.h_relocEntries, h*.h_relocCount * sizeof(RelocEntry_t));
    fi;
corp;

/*
 * findLabel1 - actual routine to find a label at the given position in the
 *	current hunk.
 */

proc findLabel1(register *Hunk_t h; register ulong addr)*char:
    register *DefList_t dfl;

    dfl := h*.h_lastLabel;
    while dfl ~= nil and dfl*.dfl_offset < addr do
	dfl := dfl*.dfl_next;
    od;
    if dfl = nil or dfl*.dfl_offset ~= addr then
	h*.h_lastLabel := dfl;
	nil
    else
	h*.h_lastLabel := dfl*.dfl_next;
	dfl*.dfl_name
    fi
corp;

/*
 * findReloc - search the relocation information for the hunk to see if there
 *	is a relocatable reference from the given address.
 */

proc findReloc(*Hunk_t h; register ulong addr, refAt;
	       *ulong pTrueAddr, pHunkNumber)*char:
    register *RelocEntry_t rle;
    register *DefList_t dfl, prev;
    register *char p @ dfl;
    register byte b;

    rle := h*.h_lastReloc;
    while rle*.rle_offset < refAt do
	rle := rle + sizeof(RelocEntry_t);
    od;
    h*.h_lastReloc := rle;
    if rle*.rle_offset = refAt then
	if rle*.rle_u.rle_hunk ~= nil then
	    if rle*.rle_isRef then
		pTrueAddr* := addr;
		pHunkNumber* := NO_HUNK;
		rle*.rle_u.rle_ref*.rfl_name
	    else
		dfl := rle*.rle_u.rle_hunk*.hkl_symbols;
		pHunkNumber* := rle*.rle_u.rle_hunk*.hkl_hunkNumber;
		prev := nil;
		while dfl ~= nil and dfl*.dfl_offset <= addr do
		    prev := dfl;
		    dfl := dfl*.dfl_next;
		od;
		if prev = nil then
		    pTrueAddr* := addr;
		    if h*.h_labelsOK and
			rle*.rle_u.rle_hunk*.hkl_hunkNumber = h*.h_hunkNumber
		    then
			buildLabel(&HunkNameBuffer[0], addr);
			&HunkNameBuffer[0]
		    else
			p := &HunkNameBuffer[16];
			p* := '\e';
			p := p - sizeof(char);
			p* := '>';
			p := p - sizeof(char);
			while
			    b := addr & 0xf;
			    p* := if b >= 10 then b - 10 + 'a' else b + '0' fi;
			    p := p - sizeof(char);
			    addr := addr >> 4;
			    addr ~= 0
			do
			od;
			p* := 'x';
			p := p - sizeof(char);
			p* := '0';
			p := p - sizeof(char);
			p* := ':';
			p := p - sizeof(char);
			addr := rle*.rle_u.rle_hunk*.hkl_hunkNumber;
			if addr > 999 then
			    addr := 999;
			fi;
			while
			    p* := addr % 10 + '0';
			    p := p - sizeof(char);
			    addr := addr / 10;
			    addr ~= 0
			do
			od;
			p* := '<';
			p
		    fi
		else
		    pTrueAddr* := prev*.dfl_offset;
		    prev*.dfl_name
		fi
	    fi
	else
	    pTrueAddr* := addr;
	    pHunkNumber* := NO_HUNK;
	    "<reloc>"
	fi
    else
	nil
    fi
corp;

/*
 * findRelocOrDef - find a relocatable reference for the given address, made
 *	at the given offset. If none is found, check for an address define
 *	for the current hunk.
 */

proc findRelocOrDef(*Hunk_t h; ulong addr, refAt;
		    *ulong pTrueAddr, pHunkNumber)*char:
    *char label;
    register *DefList_t dfl;

    label := findReloc(h, addr, refAt, pTrueAddr, pHunkNumber);
    if label = nil then
	dfl := h*.h_defList;
	while dfl ~= nil and dfl*.dfl_offset ~= addr do
	    dfl := dfl*.dfl_next;
	od;
	if dfl ~= nil then
	    pTrueAddr* := dfl*.dfl_offset;
	    pHunkNumber* := h*.h_hunkNumber;
	    label := dfl*.dfl_name
	else
	    label := nil;
	fi;
    fi;
    label
corp;

/*
 * isLabel1 - bottom level routine to see if the given offset in the hunk
 *	has been used as a label.
 */

proc isLabel1(*Hunk_t h; ulong offset)bool:

    if offset < h*.h_bytesLength then
	(h*.h_labelMap + offset / 8)* & (1 << (offset % 8)) ~= 0
    else
	false
    fi
corp;

/*
 * isBranch1 - bottom level routine to see if the given offset in the hunk
 *	has been used as a branch target.
 */

proc isBranch1(*Hunk_t h; ulong offset)bool:

    if offset < h*.h_bytesLength then
	(h*.h_branchMap + offset / 8)* & (1 << (offset % 8)) ~= 0
    else
	false
    fi
corp;

/*
 * branchTo1 - bottom level routine to mark an offset as a branch target.
 */

proc branchTo1(*Hunk_t h; ulong offset)void:
    register *byte p;

    if offset < h*.h_bytesLength then
	p := h*.h_branchMap + offset / 8;
	p* := p* | (1 << (offset % 8));
	p := h*.h_labelMap + offset / 8;
	p* := p* | (1 << (offset % 8));
    fi;
corp;

/*
 * labelAt1 - bottom level routine to mark an offset as used as a label.
 */

proc labelAt1(*Hunk_t h; ulong offset)void:
    *byte p;

    if offset < h*.h_bytesLength then
	p := h*.h_labelMap + offset / 8;
	p* := p* | (1 << (offset % 8));
    fi;
corp;

/*
 * farLabel - a JMP or JSR to a hunk not seen yet. Remember the label.
 */

proc farLabel(ulong addr, hunkNumber; bool isBranch)void:
    register *HunkList_t hkl;
    register *LabelList_t lal;

    hkl := HunkList;
    while hkl ~= nil and hkl*.hkl_hunkNumber ~= hunkNumber do
	hkl := hkl*.hkl_next;
    od;
    if hkl = nil then
	hkl := AllocMem(sizeof(HunkList_t), 0);
	if hkl ~= nil then
	    hkl*.hkl_next := HunkList;
	    hkl*.hkl_symbols := nil;
	    hkl*.hkl_hunkNumber := hunkNumber;
	    hkl*.hkl_labelList := nil;
	    HunkList := hkl;
	fi;
    fi;
    if hkl ~= nil then
	lal := hkl*.hkl_labelList;
	while lal ~= nil and lal*.lal_offset ~= addr do
	    lal := lal*.lal_next;
	od;
	if lal = nil then
	    lal := AllocMem(sizeof(LabelList_t), 0);
	    if lal ~= nil then
		lal*.lal_next := hkl*.hkl_labelList;
		lal*.lal_offset := addr;
		lal*.lal_isBranch := isBranch;
		hkl*.hkl_labelList := lal;
	    fi;
	else
	    if isBranch and not lal*.lal_isBranch then
		lal*.lal_isBranch := true;
	    fi;
	fi;
    fi;
corp;

/*
 * makeLabelMap - allocate a label map for the current hunk.
 */

proc makeLabelMap(register *Hunk_t h)void:
    register *byte bp;
    register *HunkList_t hkl @ bp;
    register *LabelList_t lal;
    register ulong length, i;

    length := (h*.h_bytesLength + 7) / 8;
    h*.h_mapLength := length;
    bp := AllocMem(length, 0);
    h*.h_labelMap := bp;
    if bp ~= nil then
	for i from length - 1 downto 0 do
	    bp* := 0;
	    bp := bp + 1;
	od;
	bp := AllocMem(length, 0);
	h*.h_branchMap := bp;
	if bp ~= nil then
	    h*.h_labelsOK := true;
	    for i from length - 1 downto 0 do
		bp* := 0;
		bp := bp + 1;
	    od;
	    hkl := HunkList;
	    while hkl ~= nil and hkl*.hkl_hunkNumber ~= h*.h_hunkNumber do
		hkl := hkl*.hkl_next;
	    od;
	    if hkl ~= nil then
		lal := hkl*.hkl_labelList;
		while lal ~= nil do
		    if lal*.lal_isBranch then
			branchTo1(h, lal*.lal_offset);
		    else
			labelAt1(h, lal*.lal_offset);
		    fi;
		    lal := lal*.lal_next;
		od;
		freeLabelList(hkl*.hkl_labelList);
		hkl*.hkl_labelList := nil;
	    fi;
	else
	    FreeMem(h*.h_labelMap, length);
	    fileError("can't allocate branch map - labels supressed\n");
	fi;
    else
	fileError("can't allocate label map - labels supressed\n");
    fi;
corp;

/*
 * symbolInit - start up processing here.
 */

proc symbolInit()void:

    FakeReloc.rle_offset := 0xffffffff;
    FakeReloc.rle_u.rle_hunk := nil;
    FakeReloc.rle_isRef := false;
    HunkList := nil;
corp;

/*
 * symbolTerm - shut down symbolic processing for this file.
 */

proc symbolTerm()void:
    register *HunkList_t hkl, hklTemp;

    hkl := HunkList;
    while hkl ~= nil do
	freeDefList(hkl*.hkl_symbols);
	freeLabelList(hkl*.hkl_labelList);
	hklTemp := hkl;
	hkl := hkl*.hkl_next;
	FreeMem(hklTemp, sizeof(HunkList_t));
    od;
corp;
